# README for utils/config directory

**This information is of value to developers only.**

## Basics:

Sripts below this directory are run to generate content in `upcxx_config.hpp`
and `gasnet.{codemode}.mak`.  To be run, they must be named in either the
`UPCXX_CONFIG_SCRIPTS` or `GASNET_CONFIG_SCRIPTS` variable in `bld/config.mak`.

The interface is simple: inputs via environment (below) and output on `stdout`.
Everything on `stdout` is sent to the corresponding config file.

All `stderr` is provided to the user. Exiting non-zero stops the build.

Scripts are run after GASNet has been built.  This should allow a very wide
variety of tests to be constructed.  However, be aware that it is not (in
general) possible to execute test codes when cross-compiling.

Scripts should assume multiple instances will be running concurrently for
different builds of libupcxx.a, but with distinct working directories.

Scripts should remove any temporary files they create.  The prefix `conftest`
is preferred for any temporary files (source, object, etc.) and it is
permissible to overwrite any pre-exiting files with that prefix (perhaps from
other probes that failed to cleanup).

There is a distinct `upcxx_config.hpp` header for each and every `libupcxx.a`.
Scripts generating these headers are run in the corresponding `bld/upcxx.*/`
build directory.

There are only two makefile fragments: `gasnet.opt.mak` and `gasnet.debug.mak`.
Scripts generating these makefile fragments are run in the corresponding
`bld/gasnet.{codemode}/` build directory.  Each is built with environment
variables (below) set for the SEQ build of the default conduit.

## Environment:

While GNU Make may "leak" many of its variables into the environment, scripts
should only depend on ones documented here.

The following environment variables carry information about the UPC\+\+ library
instance to be built using the resulting `upcxx_config.hpp`:

* `UPCXX_TOPSRC`: top of the source directory hierarchy
* `UPCXX_TOPBLD`: top of the build directory hierarchy
* `UPCXX_ASSERT`: either `0` or `1`
* `UPCXX_OPTLEV`: either `0` or `3`
* `UPCXX_DBGSYM`: either `0` or `1`
* `UPCXX_NETWORK`: the conduit (lowercase)
* `UPCXX_CODEMODE`: either `opt` or `debug`
* `UPCXX_THREADMODE`: either `seq` or `par`
* `UPCXX_CROSS`: the cross-configure platform name, or empty

Additionally, `UPCXX_GMAKE` is defined to the GNU Make discovered by
`configure` and can be used to extract additional configure outputs.  The
following will set the shell variable `UPCXX_CUDA_NVCC` to the corresponding
value in the top-level `Makefile` generated by `configure`:

```bash
eval $($UPCXX_GMAKE -C "$UPCXX_TOPBLD" echovar VARNAME=UPCXX_CUDA_NVCC)
```

The following each carry the corresponding values from the makefile fragment:

* `GASNET_CC`
* `GASNET_CFLAGS`
* `GASNET_OPT_CFLAGS`
* `GASNET_MISC_CFLAGS`
* `GASNET_CXX`
* `GASNET_CXXFLAGS`
* `GASNET_OPT_CXXFLAGS`
* `GASNET_MISC_CXXFLAGS`
* `GASNET_CPPFLAGS`
* `GASNET_MISC_CPPFLAGS`
* `GASNET_CXXCPPFLAGS`
* `GASNET_MISC_CXXCPPFLAGS`
* `GASNET_DEFINES`
* `GASNET_INCLUDES`
* `GASNET_LD`
* `GASNET_LDFLAGS`
* `GASNET_LIBS`
